@@ -9,9 +9,9 @@ use termcolor::WriteColor;
 
 use crate::{buffer::Buffer, cli::Theme};
 
-pub fn get_json_formatter() -> jsonxf::Formatter {
+pub fn get_json_formatter(indent_level: usize) -> jsonxf::Formatter {
     let mut fmt = jsonxf::Formatter::pretty_printer();
-    fmt.indent = String::from("    ");
+    fmt.indent = " ".repeat(indent_level);
     fmt.record_separator = String::from("\n\n");
     fmt.eager_record_separators = true;
     fmt
@@ -54,9 +54,15 @@ impl<'a> Highlighter<'a> {
     }
 
     /// Write a single piece of highlighted text.
+    /// May return a [`io::ErrorKind::Other`] when there is a problem
+    /// during highlighting.
     pub fn highlight(&mut self, text: &str) -> io::Result<()> {
         for line in LinesWithEndings::from(text) {
-            for (style, component) in self.highlighter.highlight(line, self.syntax_set) {
+            for (style, component) in self
+                .highlighter
+                .highlight_line(line, self.syntax_set)
+                .map_err(|e| io::Error::new(io::ErrorKind::Other, e))?
+            {
                 self.out.set_color(&convert_style(style))?;
                 write!(self.out, "{}", component)?;
             }
