@@ -19,7 +19,7 @@ mod vendored;
 use std::env;
 use std::fs::File;
 use std::io::{stdin, Read};
-use std::net::IpAddr;
+use std::net::{IpAddr, SocketAddr};
 use std::path::PathBuf;
 use std::process;
 use std::str::FromStr;
@@ -27,16 +27,20 @@ use std::sync::Arc;
 
 use anyhow::{anyhow, Context, Result};
 use atty::Stream;
+use cookie_store::{CookieStore, RawCookie};
+#[cfg(feature = "network-interface")]
+use network_interface::{NetworkInterface, NetworkInterfaceConfig};
 use redirect::RedirectFollower;
 use reqwest::blocking::Client;
 use reqwest::header::{
     HeaderValue, ACCEPT, ACCEPT_ENCODING, CONNECTION, CONTENT_TYPE, COOKIE, RANGE, USER_AGENT,
 };
 use reqwest::tls;
+use url::Host;
 
 use crate::auth::{Auth, DigestAuthMiddleware};
 use crate::buffer::Buffer;
-use crate::cli::{BodyType, Cli, HttpVersion, Print, Proxy, Verify};
+use crate::cli::{Cli, FormatOptions, HttpVersion, Print, Proxy, Verify};
 use crate::download::{download_file, get_file_size};
 use crate::middleware::ClientWithMiddleware;
 use crate::printer::Printer;
@@ -60,7 +64,6 @@ fn get_user_agent() -> &'static str {
 fn main() {
     let args = Cli::parse();
     let bin_name = args.bin_name.clone();
-    let url = args.url.clone();
     let native_tls = args.native_tls;
 
     match run(args) {
@@ -70,22 +73,6 @@ fn main() {
         Err(err) => {
             eprintln!("{}: error: {:?}", bin_name, err);
             let msg = err.root_cause().to_string();
-            if !native_tls && msg == "invalid dnsname" {
-                eprintln!();
-                if utils::url_requires_native_tls(&url) {
-                    eprintln!("rustls does not support HTTPS for IP addresses.");
-                } else {
-                    // Maybe we went to https://<IP> after a redirect?
-                    eprintln!(
-                        "This may happen because rustls does not support HTTPS for IP addresses."
-                    );
-                }
-                if cfg!(feature = "native-tls") {
-                    eprintln!("Try using the --native-tls flag.");
-                } else {
-                    eprintln!("Consider building with the `native-tls` feature enabled.");
-                }
-            }
             if native_tls && msg == "invalid minimum TLS version for backend" {
                 eprintln!();
                 eprintln!("Try running without the --native-tls flag.");
@@ -118,26 +105,9 @@ fn run(args: Cli) -> Result<i32> {
     let url = url_with_query(args.url, &args.request_items.query()?);
 
     let use_stdin = !(args.ignore_stdin || atty::is(Stream::Stdin) || test_pretend_term());
-    let body_type = args.request_items.body_type;
-    let body_from_request_items = args.request_items.body()?;
-    let has_request_items = !body_from_request_items.is_empty();
 
-    let body = match (use_stdin, args.raw, has_request_items) {
-        (true, None, false) => {
-            let mut buffer = Vec::new();
-            stdin().read_to_end(&mut buffer)?;
-            Body::Raw(buffer)
-        }
-        (false, Some(raw), false) => Body::Raw(raw.into_bytes()),
-        (false, None, _) => body_from_request_items,
-
-        (true, Some(_), _) => {
-            return Err(anyhow!(
-                "Request body from stdin and --raw cannot be mixed. \
-                Pass --ignore-stdin to ignore standard input."
-            ))
-        }
-        (true, _, true) => {
+    let body = if use_stdin {
+        if !args.request_items.is_body_empty() {
             if args.multipart {
                 // Multipart bodies are never "empty", so we can get here without request items
                 return Err(anyhow!("Cannot build a multipart request body from stdin"));
@@ -148,26 +118,28 @@ fn run(args: Cli) -> Result<i32> {
                 ));
             }
         }
-        (_, Some(_), true) => {
-            if args.multipart {
-                // Multipart bodies are never "empty", so we can get here without request items
-                return Err(anyhow!("Cannot build a multipart request body from --raw"));
-            } else {
-                return Err(anyhow!(
-                    "Request body (from --raw) and request data (key=value) cannot be mixed."
-                ));
-            }
+        if args.raw.is_some() {
+            return Err(anyhow!(
+                "Request body from stdin and --raw cannot be mixed. \
+                Pass --ignore-stdin to ignore standard input."
+            ));
         }
+        let mut buffer = Vec::new();
+        stdin().read_to_end(&mut buffer)?;
+        Body::Raw(buffer)
+    } else if let Some(raw) = args.raw {
+        Body::Raw(raw.into_bytes())
+    } else {
+        args.request_items.body()?
     };
 
     let method = args.method.unwrap_or_else(|| body.pick_method());
-    let timeout = args.timeout.and_then(|t| t.as_duration());
 
     let mut client = Client::builder()
         .http1_title_case_headers()
         .http2_adaptive_window(true)
         .redirect(reqwest::redirect::Policy::none())
-        .timeout(timeout)
+        .timeout(args.timeout.and_then(|t| t.as_duration()))
         .no_gzip()
         .no_deflate()
         .no_brotli();
@@ -177,7 +149,7 @@ fn run(args: Cli) -> Result<i32> {
         client = client.use_rustls_tls();
     }
 
-    if let Some(Some(tls_version)) = args.ssl {
+    if let Some(tls_version) = args.ssl.and_then(Into::into) {
         client = client
             .min_tls_version(tls_version)
             .max_tls_version(tls_version);
@@ -197,10 +169,6 @@ fn run(args: Cli) -> Result<i32> {
     #[cfg(feature = "native-tls")]
     if args.native_tls {
         client = client.use_native_tls();
-    } else if utils::url_requires_native_tls(&url) {
-        // We should be loud about this to prevent confusion
-        warn("rustls does not support HTTPS for IP addresses. native-tls will be enabled. Use --native-tls to silence this warning.");
-        client = client.use_native_tls();
     }
 
     #[cfg(not(feature = "native-tls"))]
@@ -245,7 +213,7 @@ fn run(args: Cli) -> Result<i32> {
                 })?;
 
                 client = client.tls_built_in_root_certs(false);
-                for pem in pem::parse_many(buffer) {
+                for pem in pem::parse_many(buffer)? {
                     let certificate = reqwest::Certificate::from_pem(pem::encode(&pem).as_bytes())
                         .with_context(|| {
                             format!("Failed to load the custom CA bundle: {}", path.display())
@@ -318,11 +286,43 @@ fn run(args: Cli) -> Result<i32> {
         _ => client,
     };
 
+    if let Some(name_or_ip) = &args.interface {
+        let ip_addr = if let Ok(ip_addr) = IpAddr::from_str(name_or_ip) {
+            ip_addr
+        } else {
+            #[cfg(not(feature = "network-interface"))]
+            return Err(anyhow!(
+                "This binary was built without support for binding to interfaces. Enable the `network-interface` feature."
+            ));
+
+            #[cfg(feature = "network-interface")]
+            // TODO: Directly bind to interface name once hyper/reqwest adds support for it.
+            // See https://github.com/seanmonstar/reqwest/issues/1336 and https://github.com/hyperium/hyper/pull/3076
+            NetworkInterface::show()?
+                .iter()
+                .find_map(|interface| {
+                    if &interface.name == name_or_ip {
+                        if let Some(addr) = interface.addr.first() {
+                            return Some(addr.ip());
+                        }
+                    }
+                    None
+                })
+                .with_context(|| format!("Couldn't bind to {:?}", name_or_ip))?
+        };
+
+        client = client.local_address(ip_addr);
+    }
+
+    for resolve in args.resolve {
+        client = client.resolve(&resolve.domain, SocketAddr::new(resolve.addr, 0));
+    }
+
     let client = client.build()?;
 
     let mut session = match &args.session {
         Some(name_or_path) => Some(
-            Session::load_session(&url, name_or_path.clone(), args.is_session_read_only)
+            Session::load_session(url.clone(), name_or_path.clone(), args.is_session_read_only)
                 .with_context(|| {
                     format!("couldn't load session {:?}", name_or_path.to_string_lossy())
                 })?,
@@ -332,21 +332,21 @@ fn run(args: Cli) -> Result<i32> {
 
     if let Some(ref mut s) = session {
         auth = s.auth()?;
-        for (key, value) in s.headers()?.iter() {
-            headers.entry(key).or_insert_with(|| value.clone());
-        }
+
+        headers = {
+            let mut session_headers = s.headers()?;
+            session_headers.extend(headers);
+            session_headers
+        };
         s.save_headers(&headers)?;
 
         let mut cookie_jar = cookie_jar.lock().unwrap();
-        for cookie in s.cookies() {
-            match cookie_jar.insert_raw(&cookie, &url) {
-                Ok(..) | Err(cookie_store::CookieError::Expired) => {}
-                Err(err) => return Err(err.into()),
-            }
-        }
+        *cookie_jar = CookieStore::from_cookies(s.cookies(), false)
+            .context("Failed to load cookies from session file")?;
+
         if let Some(cookie) = headers.remove(COOKIE) {
-            for cookie in cookie.to_str()?.split(';') {
-                cookie_jar.insert_raw(&cookie.parse()?, &url)?;
+            for cookie in RawCookie::split_parse(cookie.to_str()?) {
+                cookie_jar.insert_raw(&cookie?, &url)?;
             }
         }
     }
@@ -395,13 +395,15 @@ fn run(args: Cli) -> Result<i32> {
                     request_builder
                 }
             }
-            Body::Raw(body) => match body_type {
-                BodyType::Json => request_builder
-                    .header(ACCEPT, HeaderValue::from_static(JSON_ACCEPT))
-                    .header(CONTENT_TYPE, HeaderValue::from_static(JSON_CONTENT_TYPE)),
-                BodyType::Form => request_builder
-                    .header(CONTENT_TYPE, HeaderValue::from_static(FORM_CONTENT_TYPE)),
-                BodyType::Multipart => unreachable!(),
+            Body::Raw(body) => {
+                if args.form {
+                    request_builder
+                        .header(CONTENT_TYPE, HeaderValue::from_static(FORM_CONTENT_TYPE))
+                } else {
+                    request_builder
+                        .header(ACCEPT, HeaderValue::from_static(JSON_ACCEPT))
+                        .header(CONTENT_TYPE, HeaderValue::from_static(JSON_CONTENT_TYPE))
+                }
             }
             .body(body),
             Body::File {
@@ -434,7 +436,7 @@ fn run(args: Cli) -> Result<i32> {
         } else if !args.ignore_netrc {
             // I don't know if it's possible for host() to return None
             // But if it does we still want to use the default entry, if there is one
-            let host = url.host().unwrap_or(url::Host::Domain(""));
+            let host = url.host().unwrap_or(Host::Domain(""));
             if let Some(entry) = netrc::find_entry(host) {
                 auth = Auth::from_netrc(auth_type, entry);
                 save_auth_in_session = false;
@@ -489,8 +491,13 @@ fn run(args: Cli) -> Result<i32> {
             &buffer,
         ),
     };
+    let theme = args.style.unwrap_or_default();
     let pretty = args.pretty.unwrap_or_else(|| buffer.guess_pretty());
-    let mut printer = Printer::new(pretty, args.style, args.stream, buffer);
+    let format_options = args
+        .format_options
+        .iter()
+        .fold(FormatOptions::default(), FormatOptions::merge);
+    let mut printer = Printer::new(pretty, theme, args.stream, buffer, format_options);
 
     let response_charset = args.response_charset;
     let response_mime = args.response_mime.as_deref();
@@ -582,13 +589,7 @@ fn run(args: Cli) -> Result<i32> {
 
     if let Some(ref mut s) = session {
         let cookie_jar = cookie_jar.lock().unwrap();
-        s.save_cookies(
-            cookie_jar
-                .matches(&url)
-                .into_iter()
-                .map(|c| cookie_crate::Cookie::from(c.clone()))
-                .collect(),
-        );
+        s.save_cookies(cookie_jar.iter_unexpired());
         s.persist()
             .with_context(|| format!("couldn't persist session {}", s.path.display()))?;
     }
