@@ -1,6 +1,6 @@
 use std::io::{stderr, stdout, Write};
 
-use anyhow::{anyhow, Result};
+use anyhow::{anyhow, Context, Result};
 use os_display::Quotable;
 use reqwest::{tls, Method};
 use std::ffi::OsString;
@@ -182,7 +182,7 @@ pub fn translate(args: Cli) -> Result<Command> {
         cmd.arg("--key");
         cmd.arg(keyfile);
     }
-    if let Some(Some(tls_version)) = args.ssl {
+    if let Some(tls_version) = args.ssl.and_then(Into::into) {
         match tls_version {
             tls::Version::TLS_1_0 => {
                 cmd.arg("--tlsv1.0");
@@ -270,6 +270,11 @@ pub fn translate(args: Cli) -> Result<Command> {
     }
 
     let url = url_with_query(args.url, &args.request_items.query()?);
+
+    if url.as_str().contains(['[', ']', '{', '}']) {
+        cmd.opt("-g", "--globoff")
+    }
+
     cmd.arg(url.to_string());
 
     // Force ipv4/ipv6 options
@@ -279,6 +284,23 @@ pub fn translate(args: Cli) -> Result<Command> {
         _ => (),
     };
 
+    if let Some(interface) = args.interface {
+        cmd.arg("--interface");
+        cmd.arg(interface);
+    };
+
+    if !args.resolve.is_empty() {
+        let port = url
+            .port_or_known_default()
+            .with_context(|| format!("Unsupported URL scheme: '{}'", url.scheme()))?;
+
+        cmd.warn("Inferred port number in --resolve from request URL.");
+        for resolve in args.resolve {
+            cmd.arg("--resolve");
+            cmd.arg(format!("{}:{}:{}", resolve.domain, port, resolve.addr));
+        }
+    }
+
     // Payload
     for (header, value) in headers.iter() {
         cmd.opt("-H", "--header");
@@ -317,7 +339,17 @@ pub fn translate(args: Cli) -> Result<Command> {
         }
     }
 
-    if args.request_items.is_multipart() {
+    if let Some(raw) = args.raw {
+        if args.form {
+            cmd.header("content-type", FORM_CONTENT_TYPE);
+        } else {
+            cmd.header("content-type", JSON_CONTENT_TYPE);
+            cmd.header("accept", JSON_ACCEPT);
+        }
+
+        cmd.opt("-d", "--data");
+        cmd.arg(raw);
+    } else if args.request_items.is_multipart() {
         // We can't use .body() here because we can't look inside the multipart
         // form after construction and we don't want to actually read the files
         for item in args.request_items.items {
@@ -508,6 +540,10 @@ mod tests {
                 #[cfg(windows)]
                 "curl http://httpbin.org/post -H 'content-type: text/plain' --data-binary '@foo.txt'",
             ),
+            (
+                "xh http://example.com/[1-100].png?q={80,90}",
+                "curl -g 'http://example.com/[1-100].png?q={80,90}'",
+            )
         ];
         for (input, output) in expected {
             let cli = Cli::try_parse_from(input.split_whitespace()).unwrap();
