@@ -5,7 +5,7 @@ use std::path::{Path, PathBuf};
 
 use anyhow::Result;
 use reqwest::blocking::Request;
-use url::{Host, Url};
+use url::Url;
 
 pub fn unescape(text: &str, special_chars: &'static str) -> String {
     let mut out = String::new();
@@ -171,7 +171,3 @@ pub fn copy_largebuf(
         }
     }
 }
-
-pub fn url_requires_native_tls(url: &Url) -> bool {
-    url.scheme() == "https" && matches!(url.host(), Some(Host::Ipv4(..)) | Some(Host::Ipv6(..)))
-}
