@@ -10,7 +10,7 @@ use std::iter::FromIterator;
 use std::net::IpAddr;
 use std::pin::Pin;
 use std::str::FromStr;
-use std::time::Duration;
+use std::time::{Duration, SystemTime, UNIX_EPOCH};
 
 use assert_cmd::cmd::Command;
 use indoc::indoc;
@@ -827,7 +827,7 @@ fn successful_digest_auth() {
     get_command()
         .arg("--auth-type=digest")
         .arg("--auth=ahmed:12345")
-        .arg("httpbin.org/digest-auth/5/ahmed/12345")
+        .arg("httpbingo.org/digest-auth/auth/ahmed/12345")
         .assert()
         .stdout(contains("HTTP/1.1 200 OK"));
 }
@@ -838,7 +838,7 @@ fn unsuccessful_digest_auth() {
     get_command()
         .arg("--auth-type=digest")
         .arg("--auth=ahmed:wrongpass")
-        .arg("httpbin.org/digest-auth/5/ahmed/12345")
+        .arg("httpbingo.org/digest-auth/auth/ahmed/12345")
         .assert()
         .stdout(contains("HTTP/1.1 401 Unauthorized"));
 }
@@ -1145,7 +1145,7 @@ fn proxy_multiple_valid_proxies() {
         .arg("--proxy=https:socks5://127.0.0.1:8000")
         .arg("--proxy=all:http://127.0.0.1:8000")
         .arg("GET")
-        .arg("http://httpbin.org/get");
+        .arg("http://httpbingo.org/get");
 
     cmd.assert().success();
 }
@@ -1227,6 +1227,34 @@ fn cert_without_key() {
         .stderr(predicates::str::is_empty());
 }
 
+#[test]
+fn override_dns_resolution() {
+    let server = server::http(|req| async move {
+        let host = req.headers()["host"].to_str().unwrap();
+        assert!(host.starts_with("example.com"));
+
+        hyper::Response::builder()
+            .header("X-Foo", "Bar")
+            .header("Date", "N/A")
+            .header("Content-Type", "application/json")
+            .body(r#"{"hello":"world"}"#.into())
+            .unwrap()
+    });
+
+    get_command()
+        .arg("--body")
+        .arg("--resolve=example.com:127.0.0.1")
+        .arg(format!("http://example.com:{}", server.port()))
+        .assert()
+        .stdout(indoc! {r#"
+            {
+                "hello": "world"
+            }
+
+
+        "#});
+}
+
 #[cfg(feature = "online-tests")]
 #[test]
 fn use_ipv4() {
@@ -1292,18 +1320,6 @@ fn native_tls_flag_disabled() {
         .stderr(contains("built without native-tls support"));
 }
 
-#[cfg(all(not(feature = "native-tls"), feature = "online-tests"))]
-#[ignore = "https://1.1.1.1 randomly times out in CI"]
-#[test]
-fn improved_https_ip_error_no_support() {
-    get_command()
-        .arg("https://1.1.1.1")
-        .assert()
-        .failure()
-        .stderr(contains("rustls does not support"))
-        .stderr(contains("building with the `native-tls` feature"));
-}
-
 #[cfg(all(feature = "native-tls", feature = "online-tests"))]
 #[test]
 fn native_tls_works() {
@@ -1313,35 +1329,6 @@ fn native_tls_works() {
         .success();
 }
 
-#[cfg(all(feature = "native-tls", feature = "rustls", feature = "online-tests"))]
-#[ignore = "https://1.1.1.1 randomly times out in CI"]
-#[test]
-fn improved_https_ip_error_with_support() {
-    let server = server::http(|_req| async move {
-        hyper::Response::builder()
-            .status(301)
-            .header("Location", "https://1.1.1.1")
-            .body("Moved Permanently".into())
-            .unwrap()
-    });
-    get_command()
-        .args(["--follow", &server.base_url()])
-        .assert()
-        .failure()
-        .stderr(contains("rustls does not support"))
-        .stderr(contains("using the --native-tls flag"));
-}
-
-#[cfg(all(feature = "native-tls", feature = "rustls"))]
-#[test]
-fn auto_nativetls() {
-    get_command()
-        .args(["--offline", "https://1.1.1.1"])
-        .assert()
-        .success()
-        .stderr(contains("native-tls will be enabled"));
-}
-
 #[cfg(feature = "online-tests")]
 #[test]
 fn good_tls_version() {
@@ -1668,7 +1655,7 @@ fn multipart_raw() {
         .args(["--offline", "--raw=hello", "--multipart", ":"])
         .assert()
         .failure()
-        .stderr(contains("Cannot build a multipart request body from --raw"));
+        .stderr(contains("'--raw <RAW>' cannot be used with '--multipart'"));
 }
 
 #[test]
@@ -1985,7 +1972,7 @@ fn can_unset_default_headers() {
 #[test]
 fn can_unset_headers() {
     get_command()
-        .args([":", "hello:world", "goodby:world", "goodby:", "--offline"])
+        .args([":", "hello:world", "goodbye:world", "goodbye:", "--offline"])
         .assert()
         .stdout(indoc! {r#"
             GET / HTTP/1.1
@@ -2059,11 +2046,11 @@ fn named_sessions() {
                 "xh": "0.0.0"
             },
             "auth": { "type": "bearer", "raw_auth": "hello" },
-            "cookies": {
-                "cook1": { "value": "one", "path": "/" },
-                "lang": { "value": "en" }
-            },
-            "headers": {}
+            "cookies": [
+                { "name": "lang", "value": "en", "domain": "127.0.0.1" },
+                { "name": "cook1", "value": "one", "path": "/", "domain": "127.0.0.1" }
+            ],
+            "headers": []
         })
     );
 }
@@ -2101,8 +2088,12 @@ fn anonymous_sessions() {
                 "xh": "0.0.0"
             },
             "auth": { "type": "basic", "raw_auth": "me:pass" },
-            "cookies": { "cook1": { "value": "one" } },
-            "headers": { "hello": "world" }
+            "cookies": [
+                { "name": "cook1", "value": "one", "domain": "127.0.0.1" }
+            ],
+            "headers": [
+                { "name": "hello", "value": "world" }
+            ]
         })
     );
 }
@@ -2120,8 +2111,12 @@ fn anonymous_read_only_session() {
     let old_session_content = serde_json::json!({
         "__meta__": { "about": "xh session file", "xh": "0.0.0" },
         "auth": { "type": null, "raw_auth": null },
-        "cookies": { "cookie1": { "value": "one" } },
-        "headers": { "hello": "world" }
+        "cookies": [
+            { "name": "cookie1", "value": "one" }
+        ],
+        "headers": [
+            { "name": "hello", "value": "world" }
+        ]
     });
 
     std::fs::write(&session_file, old_session_content.to_string()).unwrap();
@@ -2178,12 +2173,12 @@ fn session_files_are_created_in_read_only_mode() {
                 "xh": "0.0.0"
             },
             "auth": { "type": null, "raw_auth": null },
-            "cookies": {
-                "lang": { "value": "ar" }
-            },
-            "headers": {
-                "hello": "world"
-            }
+            "cookies": [
+                { "name": "lang", "value": "ar", "domain": "127.0.0.1" }
+            ],
+            "headers": [
+                { "name": "hello", "value": "world" }
+            ]
         })
     );
 }
@@ -2212,12 +2207,12 @@ fn named_read_only_session() {
     let old_session_content = serde_json::json!({
         "__meta__": { "about": "xh session file", "xh": "0.0.0" },
         "auth": { "type": null, "raw_auth": null },
-        "cookies": {
-            "cookie1": { "value": "one" }
-        },
-        "headers": {
-            "hello": "world"
-        }
+        "cookies": [
+            { "name": "cookie1", "value": "one" }
+        ],
+        "headers": [
+            { "name": "hello", "value": "world" }
+        ]
     });
     fs::create_dir_all(path_to_session.parent().unwrap()).unwrap();
     File::create(&path_to_session).unwrap();
@@ -2240,7 +2235,6 @@ fn named_read_only_session() {
 
 #[test]
 fn expired_cookies_are_removed_from_session() {
-    use std::time::{SystemTime, UNIX_EPOCH};
     let future_timestamp = SystemTime::now()
         .duration_since(UNIX_EPOCH)
         .unwrap()
@@ -2255,27 +2249,33 @@ fn expired_cookies_are_removed_from_session() {
         serde_json::json!({
             "__meta__": { "about": "xh session file", "xh": "0.0.0" },
             "auth": { "type": null, "raw_auth": null },
-            "cookies": {
-                "expired_cookie": {
+            "cookies": [
+                {
+                    "name": "expired_cookie",
                     "value": "random_string",
-                    "expires": past_timestamp
+                    "expires": past_timestamp,
+                    "domain": "127.0.0.1"
                 },
-                "unexpired_cookie": {
+                {
+                    "name": "unexpired_cookie",
                     "value": "random_string",
-                    "expires": future_timestamp
+                    "expires": future_timestamp,
+                    "domain": "127.0.0.1"
                 },
-                "with_out_expiry": {
+                {
+                    "name": "with_out_expiry",
                     "value": "random_string",
+                    "domain": "127.0.0.1"
                 }
-            },
-            "headers": {}
+            ],
+            "headers": []
         })
         .to_string(),
     )
     .unwrap();
 
     get_command()
-        .arg(":")
+        .arg("127.0.0.1")
         .arg(format!(
             "--session={}",
             session_file.path().to_string_lossy()
@@ -2290,16 +2290,20 @@ fn expired_cookies_are_removed_from_session() {
         serde_json::json!({
             "__meta__": { "about": "xh session file", "xh": "0.0.0" },
             "auth": { "type": null, "raw_auth": null },
-            "cookies": {
-                "unexpired_cookie": {
+            "cookies": [
+                {
+                    "name": "unexpired_cookie",
                     "value": "random_string",
-                    "expires": future_timestamp
+                    "expires": future_timestamp,
+                    "domain": "127.0.0.1"
                 },
-                "with_out_expiry": {
+                {
+                    "name": "with_out_expiry",
                     "value": "random_string",
+                    "domain": "127.0.0.1"
                 }
-            },
-            "headers": {}
+            ],
+            "headers": []
         })
     );
 }
@@ -2332,11 +2336,11 @@ fn cookies_override_each_other_in_the_correct_order() {
         serde_json::json!({
             "__meta__": { "about": "xh session file", "xh": "0.0.0" },
             "auth": { "type": null, "raw_auth": null },
-            "cookies": {
-                "lang": { "value": "fr" },
-                "cook2": { "value": "three" }
-            },
-            "headers": {}
+            "cookies": [
+                { "name": "lang", "value": "fr", "domain": "127.0.0.1" },
+                { "name": "cook2", "value": "three", "domain": "127.0.0.1" }
+            ],
+            "headers": []
         })
         .to_string(),
     )
@@ -2361,12 +2365,81 @@ fn cookies_override_each_other_in_the_correct_order() {
         serde_json::json!({
             "__meta__": { "about": "xh session file", "xh": "0.0.0" },
             "auth": { "type": null, "raw_auth": null },
-            "cookies": {
-                "lang": { "value": "en" },
-                "cook1": { "value": "one" },
-                "cook2": { "value": "two" }
-            },
-            "headers": {}
+            "cookies": [
+                { "name": "lang", "value": "en", "domain": "127.0.0.1" },
+                { "name": "cook2", "value": "two", "domain": "127.0.0.1" },
+                { "name": "cook1", "value": "one", "domain": "127.0.0.1" },
+            ],
+            "headers": []
+        })
+    );
+}
+
+#[test]
+fn cookies_are_segmented_by_domain() {
+    let session_file = NamedTempFile::new().unwrap();
+
+    std::fs::write(
+        &session_file,
+        serde_json::json!({
+            "__meta__": { "about": "xh session file", "xh": "0.0.0" },
+            "auth": { "type": null, "raw_auth": null },
+            "cookies": [
+                // will be overwritten by set-cookie header from example.com
+                { "name": "lang", "value": "fi", "domain": "example.com" },
+                // will not be overwritten
+                { "name": "lang", "value": "fr", "domain": "example.org" },
+            ],
+            "headers": []
+        })
+        .to_string(),
+    )
+    .unwrap();
+
+    let server = server::http(|req| async move {
+        match req.uri().host() {
+            Some("example.com") => {
+                assert_eq!(req.headers()["cookie"].to_str().unwrap(), "lang=fi");
+                hyper::Response::builder()
+                    .header("set-cookie", "lang=en")
+                    .body("".into())
+                    .unwrap()
+            }
+            Some("example.net") => {
+                assert!(req.headers().get("cookie").is_none());
+                hyper::Response::builder()
+                    .header("set-cookie", "lang=ar")
+                    .body("".into())
+                    .unwrap()
+            }
+            _ => panic!("unknown path"),
+        }
+    });
+
+    for url in ["http://example.com", "http://example.net"] {
+        get_command()
+            .arg(url)
+            .arg(format!("--proxy=all:{}", server.base_url()))
+            .arg(format!(
+                "--session={}",
+                session_file.path().to_string_lossy()
+            ))
+            .assert()
+            .success();
+    }
+
+    let session_content = fs::read_to_string(session_file.path()).unwrap();
+    assert_eq!(
+        serde_json::from_str::<serde_json::Value>(&session_content).unwrap(),
+        serde_json::json!({
+            "__meta__": { "about": "xh session file", "xh": "0.0.0" },
+            "auth": { "type": null, "raw_auth": null },
+            "cookies": [
+                { "name": "lang", "value": "en", "domain": "example.com" },
+                { "name": "lang", "value": "fr", "domain": "example.org" },
+                { "name": "lang", "value": "ar", "domain": "example.net" }
+            ],
+            "headers": []
         })
     );
 }
@@ -2385,8 +2458,8 @@ fn basic_auth_from_session_is_used() {
         serde_json::json!({
             "__meta__": { "about": "xh session file", "xh": "0.0.0" },
             "auth": { "type": "basic", "raw_auth": "user:pass" },
-            "cookies": {},
-            "headers": {}
+            "cookies": [],
+            "headers": []
         })
         .to_string(),
     )
@@ -2417,8 +2490,8 @@ fn bearer_auth_from_session_is_used() {
         serde_json::json!({
             "__meta__": { "about": "xh session file", "xh": "0.0.0" },
             "auth": { "type": "bearer", "raw_auth": "secret-token" },
-            "cookies": {},
-            "headers": {}
+            "cookies": [],
+            "headers": []
         })
         .to_string(),
     )
@@ -2474,10 +2547,158 @@ fn auth_netrc_is_not_persisted_in_session() {
                 "xh": "0.0.0"
             },
             "auth": { "type": null, "raw_auth": null },
-            "cookies": {},
-            "headers": {
-                "hello": "world"
-            }
+            "cookies": [],
+            "headers": [
+                { "name": "hello", "value": "world" }
+            ]
+        })
+    );
+}
+
+#[test]
+fn multiple_headers_with_same_key_in_session() {
+    let server = server::http(|req| async move {
+        use reqwest::header::HeaderValue;
+        assert_eq!(
+            req.headers()
+                .get_all("hello")
+                .into_iter()
+                .collect::<Vec<_>>(),
+            [
+                &HeaderValue::from_static("world"),
+                &HeaderValue::from_static("people")
+            ]
+        );
+        hyper::Response::default()
+    });
+
+    let session_file = NamedTempFile::new().unwrap();
+
+    std::fs::write(
+        &session_file,
+        serde_json::json!({
+            "__meta__": { "about": "xh session file", "xh": "0.0.0" },
+            "auth": {},
+            "cookies": [],
+            "headers": [
+                { "name": "hello", "value": "world" },
+                { "name": "hello", "value": "people" },
+            ]
+        })
+        .to_string(),
+    )
+    .unwrap();
+
+    get_command()
+        .arg(server.base_url())
+        .arg(format!(
+            "--session={}",
+            session_file.path().to_string_lossy()
+        ))
+        .arg("--no-check-status")
+        .assert()
+        .success();
+}
+
+#[test]
+fn headers_from_session_are_overwritten() {
+    let server = server::http(|req| async move {
+        use reqwest::header::HeaderValue;
+        assert_eq!(
+            req.headers()
+                .get_all("hello")
+                .into_iter()
+                .collect::<Vec<_>>(),
+            [&HeaderValue::from_static("people")]
+        );
+        hyper::Response::default()
+    });
+
+    let session_file = NamedTempFile::new().unwrap();
+
+    std::fs::write(
+        &session_file,
+        serde_json::json!({
+            "__meta__": { "about": "xh session file", "xh": "0.0.0" },
+            "auth": {},
+            "cookies": [],
+            "headers": [
+                { "name": "hello", "value": "world" },
+            ]
+        })
+        .to_string(),
+    )
+    .unwrap();
+
+    get_command()
+        .arg(server.base_url())
+        .arg(format!(
+            "--session={}",
+            session_file.path().to_string_lossy()
+        ))
+        .arg("--no-check-status")
+        .arg("hello:people")
+        .assert()
+        .success();
+}
+
+#[test]
+fn old_session_format_is_automatically_migrated() {
+    let server = server::http(|req| async move {
+        assert_eq!(req.headers()["hello"], "world");
+        hyper::Response::default()
+    });
+
+    let session_file = NamedTempFile::new().unwrap();
+
+    let future_timestamp = SystemTime::now()
+        .duration_since(UNIX_EPOCH)
+        .unwrap()
+        .as_secs()
+        + 1000;
+
+    std::fs::write(
+        &session_file,
+        serde_json::json!({
+            "__meta__": { "about": "xh session file", "xh": "0.0.0" },
+            "auth": {},
+            "cookies": {
+                "lang": { "value": "en", "expires": future_timestamp, "path": "/", "secure": false },
+            },
+            "headers": { "hello": "world" }
+        })
+        .to_string(),
+    )
+    .unwrap();
+
+    get_command()
+        .arg(server.base_url())
+        .arg(format!(
+            "--session={}",
+            session_file.path().to_string_lossy()
+        ))
+        .assert()
+        .success();
+
+    let session_content = fs::read_to_string(session_file).unwrap();
+    assert_eq!(
+        serde_json::from_str::<serde_json::Value>(&session_content).unwrap(),
+        serde_json::json!({
+            "__meta__": { "about": "xh session file", "xh": "0.0.0" },
+            "auth": { "type": null, "raw_auth": null },
+            "cookies": [
+                {
+                    "name": "lang",
+                    "value": "en",
+                    "expires": future_timestamp,
+                    "path": "/",
+                    "secure": false,
+                    "domain": "127.0.0.1"
+                },
+            ],
+            "headers": [
+                { "name": "hello", "value": "world" }
+            ]
         })
     );
 }
@@ -3182,7 +3403,7 @@ fn digest_auth_with_response_meta() {
         .arg("--auth-type=digest")
         .arg("--auth=ahmed:12345")
         .arg("-vv")
-        .arg("httpbin.org/digest-auth/5/ahmed/12345")
+        .arg("httpbingo.org/digest-auth/auth/ahmed/12345")
         .assert()
         .stdout(contains("Elapsed time: ").count(2));
 }
@@ -3194,3 +3415,85 @@ fn non_get_redirect_translation_warning() {
         .assert()
         .stderr(contains("Using a combination of -X/--request and -L/--location which may cause unintended side effects."));
 }
+
+#[test]
+fn custom_json_indent_level() {
+    let server = server::http(|_req| async move {
+        hyper::Response::builder()
+            .header("content-type", "application/json")
+            .body(r#"{"hello":"world"}"#.into())
+            .unwrap()
+    });
+    get_command()
+        .args([
+            "--print=b",
+            "--format-options=json.indent:2",
+            &server.base_url(),
+        ])
+        .assert()
+        .stdout(indoc! {r#"
+            {
+              "hello": "world"
+            }
+
+
+        "#});
+}
+
+#[test]
+fn unsorted_headers() {
+    let server = server::http(|_req| async move {
+        hyper::Response::builder()
+            .header("X-Foo", "Bar")
+            .header("Date", "N/A")
+            .header("Content-Type", "application/json")
+            .body(r#"{"hello":"world"}"#.into())
+            .unwrap()
+    });
+    get_command()
+        .args(["--format-options=headers.sort:false", &server.base_url()])
+        .assert()
+        .stdout(indoc! {r#"
+            HTTP/1.1 200 OK
+            X-Foo: Bar
+            Date: N/A
+            Content-Type: application/json
+            Content-Length: 17
+
+            {
+                "hello": "world"
+            }
+
+
+        "#});
+}
+
+#[test]
+fn multiple_format_options_are_merged() {
+    let server = server::http(|_req| async move {
+        hyper::Response::builder()
+            .header("X-Foo", "Bar")
+            .header("Date", "N/A")
+            .header("Content-Type", "application/json")
+            .body(r#"{"hello":"world"}"#.into())
+            .unwrap()
+    });
+    get_command()
+        .arg("--format-options=json.indent:2,json.indent:8")
+        .arg("--format-options=headers.sort:false")
+        .arg(&server.base_url())
+        .assert()
+        .stdout(indoc! {r#"
+            HTTP/1.1 200 OK
+            X-Foo: Bar
+            Date: N/A
+            Content-Type: application/json
+            Content-Length: 17
+
+            {
+                    "hello": "world"
+            }
+
+
+        "#});
+}
